#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.

import importlib.resources
import pathlib
import tempfile
import textwrap
import tomllib
import unittest

import spiki
from spiki.plugin import Phase
from spiki.renderer import Renderer
from spiki.visitor import Visitor


class VisitorTests(unittest.TestCase):

    def test_example_atomic(self):
        plugin_types = [
            "spiki.plugins.bootstrapper:Bootstrapper",
            "spiki.plugins.writer:Writer",
        ]
        examples = importlib.resources.files("spiki.examples")
        witness = []
        with (
            tempfile.TemporaryDirectory() as output_name,
            Visitor(*plugin_types) as visitor,
        ):
            visitor.options = dict(
                output=pathlib.Path(output_name).resolve(),
                paths=[examples.joinpath("atomic")],
            )

            self.assertFalse(visitor.state)
            for change in visitor.walk(*visitor.options["paths"]):
                witness.append(change)

            output_path = pathlib.Path(output_name)
            files = list(output_path.glob("*.py"))

            self.assertEqual(len(files), 1)
            text = files[0].read_text()

            pyz = output_path.with_suffix(".pyz")
            self.assertTrue(pyz.exists())

        self.assertEqual(len(visitor.state), 1, visitor.state)
        path = list(visitor.state)[0]
        self.assertEqual("__main__.py", path.name)

        check = visitor.state[path].text
        self.assertIsInstance(check, str)
        self.assertIn("import argparse", check)
        self.assertIn("class Bootstrapper", check)
        self.assertIn("def main(args):", check)
        self.assertEqual(check.count("run()"), 2)

        self.assertEqual(len([i for i in witness if i.phase == Phase.EXTEND]), 1)
        self.assertEqual(files[0].name, "__main__.py")
        self.assertEqual(text, check)



#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.


import argparse
import logging
import os.path
from pathlib import Path
import shutil
import sys

from spiki.visitor import Visitor

from spiki.plugin import Phase


default_plugin_types = [
    "spiki.plugins.finder:Finder",
    "spiki.plugins.loader:Loader",
    "spiki.plugins.bootstrapper:Bootstrapper",
    "spiki.plugins.writer:Writer",
]


def setup_logger(level=logging.INFO):
    logging.basicConfig(level=level)
    root_logger = logging.getLogger()
    for handler in root_logger.handlers:
        handler.setFormatter(
            logging.Formatter(
                fmt="{asctime}|{levelname:>8}|{phase.name:^8}| {name:<16}| {path!s:<72}| {message}",
                datefmt=None, style='{',
                defaults=dict(phase=Phase.CONFIG, path="")
            )
        )


def main(args):
    level = logging.DEBUG if args.debug else logging.INFO
    setup_logger(level=level)
    logger = logging.getLogger("spiki")
    args.output.mkdir(parents=True, exist_ok=True)

    plugin_types = args.plugin or default_plugin_types
    # TODO: A file path (.toml) is a plugin input
    with Visitor(*plugin_types, **vars(args)) as visitor:
        for n, change in enumerate(visitor.walk(*args.paths)):
            pass

    logger.info(f"Completed {n} actions", extra=dict(phase=Phase.REPORT))
    return 0


def parser():
    default_path = Path.cwd().joinpath("output").resolve()
    rv = argparse.ArgumentParser(usage=__doc__, fromfile_prefix_chars="=")
    rv.add_argument("paths", nargs="+", type=Path, help="Specify file paths")
    rv.add_argument("-O", "--output", type=Path, default=default_path, help=f"Specify output directory [{default_path}]")
    rv.add_argument("--plugin", action="append", help=f"Specify plugin list {default_plugin_types}")
    rv.add_argument("--debug", action="store_true", default=False, help=f"Display debug logs")
    rv.convert_arg_line_to_args = lambda x: x.split()
    return rv


def run():
    p = parser()
    args, res = p.parse_known_args()
    rv = main(args)
    sys.exit(rv)


if __name__ == "__main__":
    run()



#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.

"""
https://stackoverflow.com/questions/61679282/create-web-server-using-only-standard-library
https://html.spec.whatwg.org/multipage/
"""

from collections import ChainMap
from collections.abc import Generator
import copy
import enum
import html
import sys
import textwrap
from types import SimpleNamespace
import warnings

from spiki.speechmark import SpeechMark


class Renderer:

    class Options(enum.Enum):
        tag_mode    = ["open", "pair", "void"]
        block_wrap  = ["div", "section", "none"]
        block_site  = ["above", "below", "stripe"]

    def __init__(self, template: dict = None, *, config: dict = None):
        self.template = template or dict()
        self.state = SimpleNamespace(attrib={}, blocks=[], config=ChainMap(config or dict()))
        self.sm = SpeechMark()

    @staticmethod
    def check_config(config: dict, options: enum.Enum):
        for option in options:
            try:
                if config[option.name] not in option.value:
                    warnings.warn(f"'{config[option.name]}' is not one of {option.value}")
            except KeyError:
                continue
        return config

    def get_option(self, option: "Option"):
        rv = self.state.config.get(option.name, None)
        return rv in option.value and rv

    def gen_blocks(self, **kwargs) -> Generator[str]:
        block_wrap = self.get_option(self.Options.block_wrap)
        for n, block in enumerate(self.state.blocks):
            if block_wrap:
                yield f'<{block_wrap} id="{n:02d}">'
            block = block.format(**kwargs)
            for line in self.sm.feed(textwrap.dedent(block).strip(), terminate=True):
                yield line.replace('<li id="', f'<li id="{n:02d}-')
            self.sm.reset()
            if block_wrap:
                yield f"</{block_wrap}>"

    def gen_nodes(self, tree: dict, **kwargs) -> Generator[str]:
        attrs =  (" " + " ".join(f'{k}="{html.escape(v)}"' for k, v in self.state.attrib.items())).rstrip()
        tag_mode = self.get_option(self.Options.tag_mode)
        pool = [(node, v) for node, v in tree.items() if isinstance(v, str)]
        for node, entry in pool:
            entry = html.escape(entry.format(**kwargs))
            if tag_mode == "open":
                yield f"<{node}{attrs}>"
            elif tag_mode == "pair":
                yield f"<{node}{attrs}>{entry}</{node}>"
            elif tag_mode == "void":
                yield f"<{node}{attrs} />"

    def walk(self, tree: dict, path: list = None, context: dict = None) -> Generator[str]:
        path = path or list()
        context = context or dict()

        self.state.attrib = tree.pop("attrib", {})
        blocks = tree.pop("blocks", "")
        self.state.blocks = [blocks] if blocks and isinstance(blocks, str) else blocks
        self.state.config = self.state.config.new_child(self.check_config(tree.pop("config", {}), self.Options))

        attrs = (" " + " ".join(f'{k}="{html.escape(v)}"' for k, v in self.state.attrib.items())).rstrip()
        tag_mode = self.get_option(self.Options.tag_mode)

        try:
            tag = next(i for i in reversed(path) if isinstance(i, str))
            if any(i for i in tree.values() if isinstance(i, str)):
                params = ""
            else:
                params = attrs

            if tag_mode in ["open", "pair"]:
                yield f"<{tag}{params}>"
            elif tag_mode == "void":
                yield f"<{tag}{params} />"
        except StopIteration:
            pass

        block_site = self.get_option(self.Options.block_site)
        if block_site == "above":
            yield from self.gen_blocks(**context)
            yield from self.gen_nodes(tree, **context)
        else:
            yield from self.gen_nodes(tree, **context)
            yield from self.gen_blocks(**context)

        pool = [(k, v) for k, v in tree.items() if isinstance(v, list)]
        for node, entry in pool:
            for n, item in enumerate(entry):
                yield from self.walk(item, path=path + [node, n], context=context)

        pool = [(k, v) for k, v in tree.items() if isinstance(v, dict)]
        for node, entry in pool:
            yield from self.walk(entry, path=path + [node], context=context)

        try:
            tag = next(i for i in reversed(path) if isinstance(i, str))
            if tag_mode == "pair":
                yield f"</{tag}>"
        except StopIteration:
            pass

        self.state.config.maps.pop(0)

    def serialize(self, template: dict = None) -> str:
        self.template.update(template or dict())
        context = copy.deepcopy(self.template)
        tree = context.pop("doc", dict())
        return "\n".join(filter(None, self.walk(tree, path=[], context=context)))



#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.

import textwrap
import tomllib
import unittest

from spiki.renderer import Renderer


class RendererTests(unittest.TestCase):

    def test_config_option(self):
        toml = textwrap.dedent("""
        [doc]
        config = {tag_mode = "invalid"}
        """)
        template = tomllib.loads(toml)
        r = Renderer()
        with self.assertWarns(UserWarning) as context:
            rv = r.serialize(template)
        self.assertFalse(rv)
        self.assertIn("invalid", format(context.warning))

    def test_config_scope(self):
        toml = textwrap.dedent("""
        [doc.html.body.header.nav]
        attrib = {popovertarget = "menu"}
        config = {tag_mode = "pair"}
        button = "Click for Menu"
        """).strip()
        template = tomllib.loads(toml)
        rv = Renderer().serialize(template)
        self.assertIn("<nav>", rv)
        self.assertIn('<button popovertarget="menu">', rv)

    def test_head(self):
        test = "Essential head content"
        goal = textwrap.dedent(f"""
        <!doctype html>
        <html lang="en">
        <head>
        <title>{test}</title>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0" />
        <meta http-equiv="X-UA-Compatible" content="ie=edge">
        </head>
        </html>
        """).strip()

        toml = textwrap.dedent("""
        [doc]
        config = {tag_mode = "open"}

        "!doctype html" = ""

        [doc.html]
        config = {tag_mode = "pair"}
        attrib = {lang = "en"}

        [doc.html.head]
        title = ""

        [[doc.html.head.meta]]
        config = {tag_mode = "open"}
        attrib = {charset = "UTF-8"}

        [[doc.html.head.meta]]
        config = {tag_mode = "void"}
        attrib = {name = "viewport", content = "width=device-width, initial-scale=1.0"}

        [[doc.html.head.meta]]
        config = {tag_mode = "open"}
        attrib = {http-equiv = "X-UA-Compatible", content = "ie=edge"}

        """)

        template = tomllib.loads(toml)
        template["doc"]["html"]["head"]["title"] = test
        rv = Renderer().serialize(template)
        self.assertEqual(rv, goal, template)

    def test_blocks(self):
        test = "Multiple speech blocks"
        toml = textwrap.dedent("""
        [doc]

        [doc.html]
        config = {tag_mode = "pair"}

        [doc.html.head]
        title = ""

        [doc.html.body]
        blocks = [
            '''
            <STAFF.proposing#3> What would you like sir? We have some very good fish today.
                1. Order the Beef Wellington
                2. Go for the Cottage Pie
                3. Try the Dover Sole

            ''',
            '''
            <GUEST.offering> Give me a another minute or two, would you?
            ''',
            '''
            <STAFF.clarifying> Certainly, sir.
            '''
        ]

        """)
        template = tomllib.loads(toml)
        template["doc"]["html"]["head"]["title"] = test
        rv = Renderer().serialize(template)
        self.assertTrue(rv.startswith("<html>"))
        self.assertTrue(rv.endswith("</html>"))
        self.assertIn("<body>", rv)
        self.assertIn("</body>", rv)
        self.assertEqual(rv.count("<blockquote"), 3)
        self.assertEqual(rv.count("</blockquote>"), 3)
        block = rv[rv.index("<blockquote"):rv.index("</blockquote>")]
        for tag in [
            "<cite", "</cite>", "<ol", '<li id="00-1"', '<li id="00-2"', '<li id="00-3"', "</ol>",
        ]:
            self.assertIn(tag, block)

    def test_block_wrap(self):
        test = "Div-wrapped speech blocks"
        toml = textwrap.dedent("""
        [doc]

        [doc.html]
        config = {tag_mode = "pair"}

        [doc.html.head]
        title = ""

        [doc.html.body]
        config = {tag_mode = "pair", block_wrap = "div"}
        blocks = [
            '''
            <STAFF.proposing#3> What would you like sir? We have some very good fish today.
                1. Order the Beef Wellington
                2. Go for the Cottage Pie
                3. Try the Dover Sole

            ''',
            '''
            <GUEST.offering> Give me a another minute or two, would you?
            ''',
            '''
            <STAFF.clarifying> Certainly, sir.
            '''
        ]

        """)
        template = tomllib.loads(toml)
        template["doc"]["html"]["head"]["title"] = test
        rv = Renderer().serialize(template)
        self.assertTrue(rv.startswith("<html>"))
        self.assertTrue(rv.endswith("</html>"))
        self.assertIn("<body>", rv)
        self.assertIn("</body>", rv)
        self.assertEqual(rv.count("<div"), 3)
        self.assertEqual(rv.count("</div>"), 3)
        self.assertEqual(rv.count("<blockquote"), 3)
        self.assertEqual(rv.count("</blockquote>"), 3)
        block = rv[rv.index("<div"):rv.index("</div>")]
        for tag in [
            "<cite", "</cite>", "<ol", '<li id="00-1"', '<li id="00-2"', '<li id="00-3"', "</ol>",
        ]:
            self.assertIn(tag, block)

    def test_context(self):
        test = "Test variable substitution & escaping"
        toml = textwrap.dedent("""
        [metadata]

        [doc]

        [doc.html]
        config = {tag_mode = "pair"}

        [doc.html.head]
        title = "{metadata[title]}"

        [doc.html.body]
        blocks = ["It's vital to {metadata[title]}."]

        """)
        template = tomllib.loads(toml)
        template["metadata"]["title"] = test
        rv = Renderer().serialize(template)
        self.assertTrue(rv.startswith("<html>"))
        self.assertTrue(rv.endswith("</html>"))
        self.assertIn("<head>", rv)
        self.assertIn("</head>", rv)
        self.assertIn("<title>Test variable substitution &amp; escaping</title>", rv)
        self.assertIn("It's vital to Test variable substitution &amp; escaping.", rv)

    def test_definition_list(self):
        test = "Test variable substitution & escaping"
        toml = textwrap.dedent("""
        [metadata]

        [doc.html]
        config = {tag_mode = "pair"}

        [[doc.html.body.main.dl.div]]
        dt = "Title"
        dd = "{metadata[title]}"

        [[doc.html.body.main.dl.div]]
        dt = "Version"
        dd = "1"

        """)
        template = tomllib.loads(toml)
        template["metadata"]["title"] = test
        rv = Renderer().serialize(template)
        self.assertTrue(rv.startswith("<html>"))
        self.assertTrue(rv.endswith("</html>"))
        self.assertEqual(rv.count("<div>"), 2)
        self.assertEqual(rv.count("</div>"), 2)
        self.assertEqual(rv.count("<dt>"), 2)
        self.assertEqual(rv.count("</dt>"), 2)
        self.assertEqual(rv.count("<dd>"), 2)
        self.assertEqual(rv.count("</dd>"), 2)
